;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-align)
(require 'bad-elem)
(require 'bad-move)
(require 'bad-pos)

(cl-defmethod resize-sub ((e elem))
  (with-slots (sub w h) e
    (dolist (s sub)
      (with-slots (resize-wh (sub-w w) (sub-h h) w-max h-max) s
        (setf w-max w)
        (setf h-max h)
        (when resize-wh
          (let ((rx (car  resize-wh))
                (ry (cadr resize-wh)))
          (size s (or (and rx (+ w rx)) sub-w)
                      (or (and ry (+ h ry)) sub-h))))))))

(cl-defmethod size ((e elem) (new-w integer) &optional new-h)
  (or new-h (setq new-h new-w))
  (with-slots (movable sub
               w w-min w-max
               h h-min h-max)
      e
    (when (and movable
               (<= w-min new-w w-max)
               (<= h-min new-h h-max))
      (let ((dx (- new-w w))
            (dy (- new-h h)))
        (setf w new-w)
        (setf h new-h)
        (align-sub e dx dy)
        (resize-sub e)))))

(cl-defmethod size-min ((e elem))
  (with-slots (min-x min-y w-min h-min) e
    (pos e min-x min-y)
    (size e w-min h-min)))

(cl-defmethod size-half ((e elem))
  (with-slots (min-x min-y w-max h-max) e
    (pos e min-x min-y)
    (size e (/ w-max 2) (/ h-max 2))))

(cl-defmethod size-square ((e elem))
  (with-slots (w-max h-max) e
    (let ((small (min w-max h-max)))
      (size e small small))))

(cl-defmethod size-max ((e elem))
  (with-slots (w h min-x min-y w-max h-max) e
    (pos e min-x min-y)
    (size e w-max h-max)))

(cl-defmethod size-add ((e elem) &optional x y)
  (or x (setq x 1))
  (or y (setq y x))
  (with-slots (w h) e
    (size e (+ x w) (+ y h))))

(cl-defmethod add-col ((e elem) &optional n)
  (or n (setq n 1))
  (size-add e n 0))

(cl-defmethod add-row ((e elem) &optional n)
  (or n (setq n 1))
  (size-add e 0 n))

(provide 'bad-size)
