;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-ascii)
(require 'bad-box)
(require 'bad-caption)
(require 'bad-game)
(require 'bad-mode)
(require 'bad-paths)
(require 'bad-write)

(defclass egypt (game)
  ((name        :initform "egypt")
   (eye         :initarg :eye         :type (or null ascii)   :initform nil)
   (egypt-sub   :initarg :egypt-sub   :type (or null box)     :initform nil)
   (crack-left  :initarg :crack-left  :type (or null box)     :initform nil)
   (crack-right :initarg :crack-right :type (or null box)     :initform nil)
   (egypt-box   :initarg :egypt-box   :type (or null box)     :initform nil)))

(cl-defmethod about ((e egypt))
  (with-slots (name) e
    (message "%s" name)))

(cl-defmethod init ((e egypt))
  (with-slots (w h w-max h-max
               help screen
               eye egypt-sub crack-left crack-right egypt-box)
      e
    (setf screen (box :name "screen" :w w :h h))
    (border screen ?\s)
    (draw screen)

    (setf eye (ascii :name "eye" :align-c t :x 6 :y 2))
    (read-file eye)

    (setf help
         (caption
          :name "help"
          :text (string-join
       #3=(list #2=(format "%s help                          [i]    " #1="[h]")
                          "[a] about                     [j] adj [l]"
                          "[q] quit                          [k]    "))
          :text-alt #1#
          :w (length #2#) :h #4=(length #3#)
          :y (- h #4# 4)))
    (draw help)

    (setq egypt-sub
          (box :name "egypt-sub"
               :x 1       :y 1
               :w (- w 2) :h (- h 2)
               :c1 #1=?\\    :s1 '(#2=?^) :c2 #2#
               :s4 '(?/ #1#)              :s2 #3=`(,bad-nonsolid)
               :c4 #1#       :s3 '(#4=?_) :c3 #4#
               :spc #3#
               :resize-wh '(-2 -2)))
    (draw egypt-sub)

  (setq crack-left
    (box :name "crack-left"
         :x 3 :y 2
         :w 2 :h 1
         :c1   #1=?_ :c2  #2=?-
         :s4 '(#1#)  :s2 '(#2#)
         :c4   #1#   :c3   #2#
         :resize-wh '(nil -4)))
  (draw crack-left)

  (setq crack-right
    (box :name "crack-right"
         :x (- w 4) :y 2
         :w 2       :h 1
         :c1   #1=?_ :c2   #2=?-
         :s4 '(#1#)  :s2 '(#2#)
         :c4   #1#   :c3   #2#
         :align-rel-xy '(-4 nil)
         :resize-wh    '(nil -4)))
  (draw crack-right)

  (setq egypt-box
    (box :name "egypt-box"
         :w #1=(1+ (* 10 (/ w #2=16)))
         :h #3=(1+ (* 18 (/ h #2#)))
         :w-min #1# :h-min #3#
         :w-max w   :h-max (- h (oref help h) 4)
         :c1 #1=?_       :s1 '(#1#) :c2 #2=?\s
         :s4 #3='(?\\ ?/)           :s2 #3#
         :c4 #2#         :s3 '(?^)  :c3 #2#
         :sub (list egypt-sub crack-left crack-right eye)))
  (setf (oref egypt-box tick) (lambda () (egypt-box-tick egypt-box)))
  (draw egypt-box)
  (with-slots (w h) egypt-box ;; NOTE: OK
    (size egypt-box w h))
  (setup-bad-mode e)))

(cl-defmethod init-keys ((e egypt))
  (setq bad-mode-map (make-keymap))
  (let ((map bad-mode-map))
    (set-char-table-range (nth 1 map) t #'ignore)
    (keymap-set map "h" (lambda () (interactive) (help-toggle e)))
    (keymap-set map "a" (lambda () (interactive) (about e)))
    (keymap-set map "q" #'kill-emacs)
    (with-slots ((eb egypt-box)) e
      (keymap-set map "i" (lambda () (interactive) (add-row eb  -1) (draw e)))
      (keymap-set map "k" (lambda () (interactive) (add-row eb)     (draw e)))
      (keymap-set map "j" (lambda () (interactive) (add-col eb  -1) (draw e)))
      (keymap-set map "l" (lambda () (interactive) (add-col eb)     (draw e))))))

(cl-defmethod setup-bad-mode ((e egypt))
  (init-keys e)
  (pop-to-buffer (bad-mode-buffer))
  (delete-other-windows))

(cl-defmethod draw ((e egypt))
  (with-slots (screen egypt-box help) e
    (draw screen)
    (draw egypt-box)
    (write-over screen help)
    (write-over screen egypt-box)
    (draw-to-buf screen)))

(cl-defmethod egypt-box-tick ((b box))
  (with-slots (w h) b
    (write-index b     (* (- h 2) w)  ?\\)
    (write-index b (1+ (* (- h 1) w)) ?`)
    (unless (zerop (mod h 2))
      (write-index b (- (* h w) 1) ?`))))

(cl-defmethod run ((e egypt))
  (init e)
  (draw e))

(provide 'bad-egypt)
