;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-abc)
(require 'bad-ascii)
(require 'bad-binary)
(require 'bad-borderless)
(require 'bad-caption)
(require 'bad-cyclic)
(require 'bad-draw)
(require 'bad-game)
(require 'bad-magni-glass)
(require 'bad-mode)
(require 'bad-mode-buffer)
(require 'bad-water)
(require 'bad-write)

(defclass toronto (game)
  ((name    :initform "toronto")
   (tor     :initarg :toronto :type (or null ascii)       :initform nil)
   (inner   :initarg :inner   :type (or null magni-glass) :initform nil)
   (outer   :initarg :outer   :type (or null borderless)  :initform nil)
   (abc-var :initarg :abc-var :type (or null abc)         :initform nil)
   (water   :initarg :water   :type (or null cyclic)      :initform nil)
   (about   :initarg :about   :type (or null caption)     :initform nil)
   (dsta    :initarg :dsta    :type (or null caption)     :initform nil)
   (tmr     :initarg :tmr     :type (or null vector)      :initform nil)))

(cl-defmethod animate ((d toronto))
  (with-slots (water) d
    (toggle water "animation")
    (draw d)))

(cl-defmethod about ((d toronto))
  (with-slots (about dsta) d
    (visible about)
    (with-slots ((oh y) visible) about
      (with-slots (y) dsta
        (if visible
            (setf y (+ oh 2))
          (setf y oh)))))
  (draw d))

(cl-defmethod find ((d toronto))
  (with-slots (inner) d
    (find inner)
    (draw d)))

(cl-defmethod init-keys ((d toronto))
  (setq bad-mode-map (make-keymap))
  (let ((map bad-mode-map))
    (set-char-table-range (nth 1 map) t #'ignore)

    (keymap-set map "h" (lambda () (interactive) (help-toggle d)))
    (keymap-set map "a" (lambda () (interactive) (about       d)))
    (keymap-set map "s" (lambda () (interactive) (animate     d)))
    (keymap-set map "f" (lambda () (interactive) (find        d)))
    (keymap-set map "q" #'kill-emacs)

    (with-slots ((i inner)) d
      (keymap-set map "I" (lambda () (interactive) (make-movable i) (add-row i -1) (draw d)))
      (keymap-set map "K" (lambda () (interactive)                  (add-row i)    (draw d)))
      (keymap-set map "J" (lambda () (interactive) (make-movable i) (add-col i -1) (draw d)))
      (keymap-set map "L" (lambda () (interactive)                  (add-col i)    (draw d)))

      (keymap-set map "j" (lambda () (interactive) (left  i) (draw d)))
      (keymap-set map "i" (lambda () (interactive) (up    i) (draw d)))
      (keymap-set map "k" (lambda () (interactive) (down  i) (draw d)))
      (keymap-set map "l" (lambda () (interactive) (right i) (draw d))))))

(cl-defmethod setup-bad-mode ((d toronto))
  (init-keys d)
  (pop-to-buffer (bad-mode-buffer))
  (delete-other-windows))

(cl-defmethod init ((d toronto))
  (with-slots (w h about dsta help outer inner water tor abc-var) d
    (setf abc-var (abc))
    (init abc-var)

    (setf tor (ascii :name "toronto"))
    (read-file tor)

    (setq about (caption
                 :name "about"
                 :visible nil
                 :text #1="dorks run away in the face of a warrior"
                 :w-max w
                 :w #2=(length #1#) :h 1
                 :x (- w #2#)       :y 6))
    (draw about)

    (setf dsta (caption
                :name #1="dsta"
                :text #2=(format "%s: OK" #1#)
                :w-max w
                :w #3=(length #2#) :h 1
                :x (- w #3#)       :y 6))
    (draw dsta)

    (setf help (caption
                :name "help"
                :text (string-join
             #3=(list #2=(format "%s help   " #1="[h]")
                                "[a] about  "
                                "[s] animate"
                                "[f] find   "
                                "[q] quit   "
                                "           "
                                "    [i]    "
                                "[j] pos [l]"
                                "    [k]    "
                                "           "
                                "    [I]    "
                                "[J] adj [K]"
                                "    [L]    "))
                :text-alt #1#
                :w (length #2#) :h (length #3#)))
    (draw help)

    (setq outer (borderless  :name "outer" :w w :h h))
    (draw outer)

    (setq inner (magni-glass :name "inner" :x (- w 18) :outer outer))
    (draw inner)
    (set-area inner)

    (let ((str "The river. It has seen too much."))
      (setq water (cyclic :name "water" :src (bad-bytes str) :w w :h 4)))
    (lake water)
    (with-slots ((water-h h)) water
      (with-slots ((tor-h h)) tor
        (pos-y water (- h water-h       4))
        (pos-y tor   (- h water-h tor-h 3))))

    (draw-draw abc-var outer (- w 39) 0)
    (setup-bad-mode d)))

(cl-defmethod status-draw ((d toronto))
  (with-slots (inner dsta) d
    (with-slots (status) inner
      (with-slots (text-alt data x w w-max) dsta
        (setf text-alt status)
        (setf x (- w-max #1=(length text-alt)))
        (setf w #1#)
        (switch dsta)))))

(cl-defmethod draw ((d toronto))
  (status-draw d)
  (with-slots (outer water tor abc-var about help inner dsta) d
    (let ((dst outer))
      (draw dst)
      (write-over dst water)
      (write-over dst tor)
      (draw-draw abc-var dst)
      (write-over dst about)
      (write-over dst dsta)
      (write-over dst help)
      (write-over dst inner)
      (draw-to-buf dst))
    (goto-char 2)))

(cl-defmethod tick ((d toronto) &optional n now)
  (or n (setq n 1))
  (with-slots (water) d
    (dotimes (_ n)
      (unless now
        (draw water))
      (draw d))))

(cl-defmethod run ((d toronto))
  (init d)
  (with-slots (tmr) d
    (setf tmr (run-with-timer 0 1 #'tick d))))

(provide 'bad-toronto)
