;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-borderless)
(require 'bad-box)
(require 'bad-box-draw)
(require 'bad-mode)

(require 'bad-game)

(require 'bad-food)
(require 'bad-snake)

(defclass snake-game (game)
  ((name :initform "snake-game")
   (snk  :initarg :snk :type (or null snake) :initform nil)
   (fod  :initarg :fod :type (or null food)  :initform nil)))

(cl-defmethod run-timer ((s snake-game))
  (with-slots (tmr) s
    (when (member tmr timer-list)
      (cancel-timer tmr))
    (setf tmr (run-with-timer 0 0.12 #'uptick s))))

(cl-defmethod pause ((s snake-game))
  (with-slots (tmr) s
    (if (member tmr timer-list)
        (cancel-timer tmr)
      ;; NOTE: (timer-activate tmr) DNC
      ;;       create a new one instead
      (run-timer s))))

(cl-defmethod quit-snake-game ((s snake-game) &optional stay)
  (with-slots (tmr) s
    (cancel-timer tmr))
  (kill-buffer (bad-mode-buffer))
  (unless stay
    (kill-emacs)))

(cl-defmethod level-2 ((s snake-game))
  (with-slots (help bg field level snk) s
    (with-slots (y) help
      (setf y 10))
    (add-row field)
    (add-col field)
    (up field)
    (with-slots (w h) field
      (size field w h)
      (draw field)
      (with-slots (score-cpn max-x max-y) snk
        (with-slots ((sy y)) score-cpn
          (setf sy (1+ sy)))
        (setf max-x (1- w))
        (setf max-y (1- h))))
    (cl-incf level)
    (read-file bg "dark.txt" nil t)
    (draw s)))

(cl-defmethod init-help ((s snake-game))
  (with-slots (help) s
    (setf help (caption
                :name "help"
                :text (string-join
             #3=(list #2=(format "%s help   " #1="[h]")
                                "[s] score  "
                                "[p] pause  "
                                "[q] quit   "
                                "           "
                                "    [i]    "
                                "[j] mov [l]"
                                "    [k]    "))
                :text-alt #1#
                :y 2
                :w (length #2#) :h (length #3#)))
    (draw help)))

(cl-defmethod init ((s snake-game))
  (with-slots (bg field screen w h snk fod) s
    (init-help s)

    (setf bg (ascii :name "background"))
    (read-file bg "cat.txt" nil t)

    (draw screen)

    (with-slots (w h) screen
      (setf field (box
                :name "field"
                :x    (*  6 (/ w 16))
                :y    (* 15 (/ h 16))
                :w #1=(*  8 (/ w 16))
                :h #2=(* 12 (/ h 16))
                :spc (string-to-list "  .  ")))
      (draw field)

      (setf snk (snake))
      (with-slots (score-cpn score) snk
        (setf score-cpn (caption
                         :name "score-caption"
                         :visible nil
                         :x 0 :y (1- #2#)
                         :w 3 :h 1))

        (draw score-cpn t)
        (draw-score snk)))

    (setf fod (food))
    (init fod field)
    (new-food-pos fod snk)

    (draw s)
    (delete-other-windows)
    (setup-bad-mode s)))

(cl-defmethod setup-bad-mode ((s snake-game))
  (init-keys s)
  (bad-mode))

(cl-defmethod init-keys ((s snake-game))
  (setq bad-mode-map (make-keymap))
  (let ((map bad-mode-map))
    (set-char-table-range (nth 1 map) t #'ignore)

    (keymap-set map "h" (lambda () (interactive) (help-toggle s)))
    (keymap-set map "p" (lambda () (interactive) (pause s)))
    (keymap-set map "q" (lambda () (interactive) (quit-snake-game s)))

    (with-slots (snk) s
      (keymap-set map "s" (lambda () (interactive) (score-toggle snk)))
      (keymap-set map "i" (lambda () (interactive) (up    snk)))
      (keymap-set map "k" (lambda () (interactive) (down  snk)))
      (keymap-set map "j" (lambda () (interactive) (left  snk)))
      (keymap-set map "l" (lambda () (interactive) (right snk))))))

(cl-defmethod score-toggle ((snk snake))
  (with-slots (score-cpn) snk
    (with-slots (visible) score-cpn
      (setf visible (not visible)))))

(cl-defmethod draw ((s snake-game))
  (with-slots (help bg field screen snk fod) s
    (draw field)
    (draw snk field)
    (draw fod field)

    (write-over screen bg)
    (write-over screen field)
    (write-over screen help)
    (draw-to-buf screen)))

(cl-defmethod uptick ((s snake-game))
  (with-slots (tmr snk fod level) s
    (do-cmd snk)
    (move snk fod)

    (with-slots (score game-over) snk
      (when (and (= level 1)
                 (< 100 score))
        (level-2 s))
      (when game-over
        (cancel-timer tmr)
        (message "GAME OVER. Score: %d" score))))
  (draw s))

(cl-defmethod run ((s snake-game))
  (init s)
  (run-timer s)
  (with-slots (snk) s
    (right snk)))

(provide 'bad-snake-game)
