;;; -*- lexical-binding: t -*-

(require 'eieio)

(require 'bad-caption)
(require 'bad-food)
(require 'bad-move)
(require 'bad-pos)
(require 'bad-write)

(defclass snake (elem)
  ((name :initform "snake")
   (data :initform (list (list #1=1 #2=1)))
   (len  :initform 1)
   (x    :initform #1#)
   (y    :initform #2#)
   (score     :initarg :score     :type   integer           :initform 0)
   (score-cpn :initarg :score-cpn :type   (or null caption) :initform nil)
   (game-over :initarg :game-over :custom boolean           :initform nil)
   (cmds      :initarg :cmds      :type   list              :initform nil)
   (mx        :initarg :mx        :type   integer           :initform 0)
   (my        :initarg :my        :type   integer           :initform 0)
   (hd        :initarg :hd        :type   (or null integer) :initform nil)
   (c         :initarg :c         :type   integer           :initform ?%)))

(cl-defmethod draw-score ((s snake) &optional dst)
  (with-slots (score-cpn score) s
    (with-slots ((cw w) text) score-cpn
      (setf text (format "[%d]" score))
      (draw score-cpn t)
      (when (cl-typep dst 'elem)
        (write-over dst score-cpn)))))

(cl-defmethod draw ((src snake) &optional dst) ; NOTE: OK
  (when (cl-typep dst 'elem)
    (with-slots (data hd c max-x max-y) src
      (with-slots (w h) dst
        (or max-x (setf max-x (1- w)))
        (or max-y (setf max-y (1- h)))
        (let ((tl))
          (when hd
            (pcase-let ((`(,x ,y) (car data)))
              (write-index dst (+ x (* w y)) hd))
            (setq tl (cdr data)))
          (pcase-dolist (`(,x ,y) (or tl data))
            (write-index dst (+ x (* w y)) c)))
        (draw-score src dst)))))

(cl-defmethod do-cmd ((s snake))
  (with-slots (mx my cmds) s
    (when cmds
      (pcase-let (( `(,dmx ,dmy) (pop cmds) ))
        (when (and dmx (zerop mx))
          (setf mx dmx)
          (setf my 0))
        (when (and dmy (zerop my))
          (setf mx 0)
          (setf my dmy))))))

(cl-defmethod eat ((s snake) (f food))
  (with-slots (x y score len) s
    (with-slots ((food-x x) (food-y y)) f
      (when (and (= x food-x) (= y food-y))
        (new-food-pos f s)
        (setf score (+ score len))))))

(cl-defmethod new-food-pos ((f food) (s snake))
  (with-slots (data) s
    (with-slots (min-x max-x min-y max-y) f
      (let* ((allowed (cl-loop
                       for xi from min-x to max-x
                       with res
                       do (cl-loop
                           for yi from min-y to max-y
                           do (push (list xi yi) res))
                       finally return
                         (cl-set-difference res data :test #'equal)))
             (len (length allowed)))
        (pcase-let ((`(,x ,y) (nth (random len) allowed)))
          (pos f x y))))))

(cl-defmethod bite ((s snake))
  (with-slots (data game-over) s
    (unless (= 1 (cl-count (car data) data :test #'equal))
      (setf game-over t))))

(cl-defmethod collide ((s snake))
  (with-slots (min-x min-y x y max-x max-y game-over) s
    (when (or (and max-x (not (< min-x x max-x)))
              (and max-y (not (< min-y y max-y))))
      (setf game-over t))))

(cl-defmethod moving-p ((s snake))
  (with-slots (mx my) s
    (not (and (zerop mx) (zerop my)))))

(cl-defmethod move ((s snake) (f food))
  (with-slots (game-over data len x y mx my) s
    (when (and (not game-over) (moving-p s))
      (mov s mx my)
      (push (list x y) data)
      (bite s)
      (collide s)
      (unless (eat s f)
        (setf data (butlast data)))
      (setf len (length data)))))

(cl-defmethod up ((s snake))
  (with-slots (cmds) s
    (push '( nil -1) cmds)))

(cl-defmethod down ((s snake))
  (with-slots (cmds) s
    (push '( nil  1) cmds)))

(cl-defmethod left ((s snake))
  (with-slots (cmds) s
    (push '(-1  nil) cmds)))

(cl-defmethod right ((s snake))
  (with-slots (cmds) s
    (push '( 1  nil) cmds)))

(provide 'bad-snake)
